"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const typescript_1 = __importDefault(require("typescript"));
class Builder {
    // Debugging
    // depth = 0;
    // in() {
    //   this.depth++;
    // }
    // out() {
    //   this.depth--;
    // }
    // log(node: ts.Node) {
    //   const spaces = " ".repeat(this.depth);
    //   const kind = ts.SyntaxKind[node.kind];
    //   console.log(`${spaces}${kind} ${node.getText()}`);
    // }
    constructor(program) {
        this.program = program;
        this.relations = new Map();
        this.dump = {};
        this.checker = program.getTypeChecker();
    }
    analyze() {
        for (const sourceFile of this.program.getSourceFiles()) {
            if (!sourceFile.isDeclarationFile) {
                this.visit(sourceFile);
            }
        }
        this.relations.forEach((types, id) => {
            const type = types.length === 0 ? "any" : types.join(" | ");
            const keys = id.split(".");
            const [last] = keys.splice(keys.length - 1);
            let cont = this.dump;
            for (const key of keys) {
                if (cont[key] === undefined)
                    cont[key] = {};
                cont = cont[key];
            }
            cont[last] = type;
        });
    }
    print() {
        return (Object.entries(this.dump)
            .map(([key, value]) => `declare let ${key}: ${printType(value)};`)
            .join("\n") + "\n");
        function printType(value) {
            if (typeof value === "string")
                return value;
            return `{ \
${Object.entries(value)
                .map(([key, value]) => `${key}: ${printType(value)}`)
                .join(", ")} \
}`;
        }
    }
    relate(binding, exp) {
        const key = binding.map(bind => bind.getText()).join(".");
        if (!this.relations.has(key))
            this.relations.set(key, []);
        const type = this.inferType(exp);
        if (type !== "any" && !this.relations.get(key).includes(type))
            this.relations.get(key).push(type);
    }
    inferType(node) {
        if (typescript_1.default.isNumericLiteral(node)) {
            return "number";
        }
        if (typescript_1.default.isStringLiteralLike(node)) {
            return "string";
        }
        if (typescript_1.default.isLiteralTypeNode(node) &&
            (node.literal.kind === typescript_1.default.SyntaxKind.TrueKeyword ||
                node.literal.kind === typescript_1.default.SyntaxKind.FalseKeyword)) {
            return "boolean";
        }
        if (typescript_1.default.isIdentifier(node)) {
            const symbol = this.checker.getSymbolAtLocation(node);
            if (symbol) {
                return this.checker.typeToString(this.checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration));
            }
        }
        if (typescript_1.default.isCallExpression(node)) {
            const argTypes = node.arguments.map(arg => this.inferType(arg));
            return `((...args: [${argTypes.join(", ")}]) => any)`;
        }
        return "any";
    }
    visit(node) {
        if (typescript_1.default.isVariableDeclaration(node) &&
            node.initializer &&
            typescript_1.default.isIdentifier(node.name)) {
            if (typescript_1.default.isIdentifier(node.initializer) &&
                this.isExternal(node.initializer)) {
                this.relate([node.initializer], node.name);
            }
            else if (typescript_1.default.isPropertyAccessExpression(node.initializer)) {
                const flat = this.flattenPropertyAccess(node.initializer);
                if (flat && this.isExternal(flat[0])) {
                    this.relate(flat, node.name);
                }
            }
        }
        else if (typescript_1.default.isBinaryExpression(node)) {
            if (typescript_1.default.isIdentifier(node.left) && this.isExternal(node.left)) {
                this.relate([node.left], node.right);
            }
            else {
                if (typescript_1.default.isPropertyAccessExpression(node.left)) {
                    const flat = this.flattenPropertyAccess(node.left);
                    if (flat && this.isExternal(flat[0])) {
                        this.relate(flat, node.right);
                    }
                }
            }
            if (typescript_1.default.isIdentifier(node.right) && this.isExternal(node.right)) {
                this.relate([node.right], node.left);
            }
            else {
                if (typescript_1.default.isPropertyAccessExpression(node.right)) {
                    const flat = this.flattenPropertyAccess(node.right);
                    if (flat && this.isExternal(flat[0])) {
                        this.relate(flat, node.left);
                    }
                }
            }
        }
        else if (typescript_1.default.isCallExpression(node)) {
            if (typescript_1.default.isIdentifier(node.expression) &&
                this.isExternal(node.expression)) {
                this.relate([node.expression], node);
            }
            else if (typescript_1.default.isPropertyAccessExpression(node.expression)) {
                const flat = this.flattenPropertyAccess(node.expression);
                if (flat && this.isExternal(flat[0])) {
                    this.relate(flat, node);
                }
            }
        }
        // this.log(node);
        // this.in();
        typescript_1.default.forEachChild(node, node => this.visit(node));
        // this.out();
    }
    flattenPropertyAccess(node) {
        if (typescript_1.default.isIdentifier(node.expression)) {
            return [node.expression, node.name];
        }
        if (typescript_1.default.isPropertyAccessExpression(node.expression)) {
            const flat = this.flattenPropertyAccess(node.expression);
            if (flat === undefined)
                return undefined;
            return [...flat, node.name];
        }
        return undefined;
    }
    isExternal(node) {
        const symbol = this.checker.getSymbolAtLocation(node);
        return !symbol; // TODO no symbol means external, should check veracity
    }
}
exports.Builder = Builder;
